--[[
    This addon is a derivative of Prospector (v9) by kergoth.
    http://www.wowinterface.com/downloads/info7947-Prospector.html
    Initial credit for the idea for this addon goes to him/her.
]]--

Expeller = LibStub("AceAddon-3.0"):NewAddon("Expeller",
    "AceConsole-3.0",
    "AceEvent-3.0",
    "AceBucket-3.0"
--[===[@debug@
    ,"LibDebugLog-1.0"
--@end-debug@]===]
)
local L = LibStub("AceLocale-3.0"):GetLocale("Expeller", false)
local G = LibStub:GetLibrary("LibGratuity-3.0")
local LBU = LibStub("LibBagUtils-1.0")

-- Handle a missing LibDebugLog. It is only for debugging anyway.
if( Expeller.ToggleDebugLog ) then
    Expeller:ToggleDebugLog( false )
else
    Expeller.Debug = function() return end
end

--[[
    OnInitialize
]]--
function Expeller:OnInitialize()
    local options = {
        name    = "Expeller",
        handler = Expeller,
        type    = 'group',
        args    = {

        },
    }
    LibStub("AceConfig-3.0"):RegisterOptionsTable(
        'Expeller', options, {'expeller'})
    Expeller.button = CreateFrame(
       'Button', 'ExpellerRunButton', UIParent, "SecureActionButtonTemplate")
    Expeller.button = assert( getglobal('ExpellerRunButton'), "Unable to aquire ExpellerRunButton frame" )
end

--[[
    OnEnable
]]--
function Expeller:OnEnable()
    self.bucket_SKILL_LINES_CHANGED = self:RegisterBucketEvent(
        {"SKILL_LINES_CHANGED","CHAT_MSG_SKILL"},
        3,
        "SKILL_LINES_CHANGED"
    );
end

--[[
    SKILL_LINES_CHANGED
        http://www.wowwiki.com/Events/Skill#SKILL_LINES_CHANGED
]]--
function Expeller:SKILL_LINES_CHANGED()
    self:SkillInfo()
    if next(self.craft_skill) == nil then
        self:Debug('/No skillRank found, disable self.')
        if( self.bucket_BAG_UPDATE ) then
            self:UnregisterBucket(self.bucket_BAG_UPDATE)
            self.bucket_BAG_UPDATE  = nil
        end
        return
    else
        if( self.bucket_BAG_UPDATE == nil )  then
            self.bucket_BAG_UPDATE = self:RegisterBucketEvent({"BAG_UPDATE"}, .5, "BAG_UPDATE");
        end
        self:BAG_UPDATE('BAG_UPDATE')
    end
    self:CreateOrRetrieveMacro()
end

--[[
    CreateOrRetrieveMacro
        Serves to initalize the macro in the macro interface.
        Will attempt to find the macro Expel or create a new one.
]]--
function Expeller:CreateOrRetrieveMacro()
    local skillRank, skillName = self:SkillInfo()
    local name      = L['Expel']
    local icon      = 1
    local body      = '/click ExpellerRunButton'
    local unknown   = 1
    local perChar   = 0

    self.macroindex = GetMacroIndexByName(name)

    if not self.macroindex then
        self.macroindex = CreateMacro(name, icon, body, unknown, perChar)
        if self.macroindex then
            self:Print(L['A macro has been created. Please type /macro to access it.'])
        else
            self:Print(L['Unable to create ExpllerRunButton macro, please clear an account-wide entry to be used.'])
        end
    end
    assert( self.macroindex, "Unable to aquire a macro index." )
    return self.macroindex
end

--[[
    SetButton
        Sets the attributes of the secure button used to trigger
        the various spells
]]--
function Expeller:SetButton(spellname,bagid,slotid,link)
    self:Debug('Setting to spell<%s>, bagid<%i>, slot<%i>, item<%s>',spellname,bagid,slotid,link)
    self.button:SetAttribute('type1', 'spell')
    self.button:SetAttribute('spell1', spellname)
    if( bagid ~= nil and slotid ~= nil ) then
        self.button:SetAttribute('target-slot1', slotid)
        self.button:SetAttribute('target-bag1', bagid)
    end
end

--[[
    SetMacro
        Updates the text in the macro, technically unnecessary but it updates
        the #showtooltip line to make it nicer.
        Additionally, if itemname is nil then it will simply cast the skillname
        as per the normal ui action.
]]--
function Expeller:SetMacro(skillName,itemName)
    local comment       = ''
    local body          = "#showtooltip %s" .. "\n%s" .. "/click ExpellerRunButton"
    local name          = L['Expel']
    local icon          = 1
    local arg_unknown   = 1
    local perChar       = 0
    local spell         = self.lookup[skillName]
    local spelllink     = GetSpellLink(spell)
    local macro_index   = self:CreateOrRetrieveMacro()

    if(  itemName == nil ) then
        itemName    = spell
        comment     = '/cast ' .. spell .. '\n#'
    end

    body = body:format(itemName,comment)
    EditMacro(macro_index, name, icon, body, arg_unknown, perChar)

end

--[[
    CHAT_MSG_SKILL
        Fired when some chat messages about skills are displayed.
        http://www.wowwiki.com/Events/Skill#CHAT_MSG_SKILL
]]--
function Expeller:CHAT_MSG_SKILL()
    self:SkillInfo()
end

--[[
    BAG_UPDATE
        Fired when a bags inventory changes.
        http://www.wowwiki.com/Events/Item#BAG_UPDATE
]]--
function Expeller:BAG_UPDATE(event, bagid)
    if next(self.craft_skill) then
        self:FindStacks()
    else
        self:Debug("/No skills found abort.")

    end
end

--[[
    FindStacks
        Main method to locate items that we can perform spells on.
--]]
function Expeller:FindStacks()
    if InCombatLockdown() then
        self:Debug('/InCombatLockdown abort.')
        return
    end

    local foundAnItem = self:scanBagSimple()
    if foundAnItem ~= nil then
        self:Debug("/An item was found.")
        return
    end

    self:Debug("No item was found.")
    for skillName, skillRank in pairs(self.craft_skill) do
        if skillRank then
            self:Debug("Reseting macro to skill %s.", skillName)
            self:SetMacro(skillName)
            break
        end
    end
end

--[[
    checkMilling
        Checks the specified item in the bags to see if it is
        Millable/Prospectable based on the player's known skills.
]]--
function Expeller:checkMilling (bagid, slotid, link, count)
    local itemid  = tonumber(strmatch(link, "item:(%d+)"))
    local _, _, itemString, itemName = string.find(link, "^|c%x+|H(.+)|h%[(.+)%]|h")
    if count < 5 then
        return
    end
    for skillName, skillRank in pairs(self.craft_skill) do
        local iSkill = self.item_db[skillName][itemid]
        if  skillRank
            and iSkill
            and iSkill <= skillRank
        then
            self:Debug("Setting %s to %s", skillName, link)
            local j= {
                ['bagid']   = bagid,
                ['slotid']  = slotid,
                ['skill']   = L[skillName],
                ['item']    = itemName
            }
            self:Debug(j)
            return j
        end
    end
    return
end

--[[
    checkDisenchanting
        Checks the specified item in the bags to see if it is disenchantable.
]]--
function Expeller:checkDisenchanting(bagid, slotid, link)
    if self.craft_skill[L['Enchanting']] then
        local spell = self.lookup[L['Enchanting']]
        local itemName, _, iRarity, _, _, sType, sSubType, _, sEquipLoc = GetItemInfo(link)
        -- aim the tooltip lookup library.
        G:SetBagItem(bagid,slotid)
        if  iRarity == 2    -- green
            and sEquipLoc   -- equippable
            and (sType == 'Armor' or sType == 'Weapon')
            and G:NumLines() > 0      -- tooltip isn't blank
            and G:Find(ITEM_BIND_ON_EQUIP, 1, 3 ) -- boe
            and not G:Find(ITEM_SOULBOUND, 1, 3 ) -- not soulbound
        then
            return {
                ['bagid']   = bagid,
                ['slotid']  = slotid,
                ['skill']   = L['Enchanting'],
                ['item']    = itemName
            }
        end
    end
    return
end

--[[
    ChangeButtonTarget
        Handles the re-init of the secure button and the macro.
]]--
function Expeller:ChangeButtonTarget( found )
    self:Debug("Found a target for, ", found.skill, 'in bag #', found.bagid, ' slot #', found.slotid)
    self:SetButton(self.lookup[found.skill],found.bagid,found.slotid)
    self:SetMacro(found.skill,found.item)
end

--[[
    scanBag
        Actually performs the scan of the bags.
        This is a subroutine so that we can exit quickly when we find something of note.
--]]
function Expeller:scanBagSimple()
    for bagid,slotid,link in LBU:Iterate("BAGS") do
        local found
        local _, count, lock = GetContainerItemInfo(bagid, slotid)
        if not lock and link then
            found = self:checkMilling(bagid,slotid,link,count) or self:checkDisenchanting(bagid,slotid,link)
            self:Debug(found)
            if found ~= nil then
                self:Debug("Found a valid item, leaving early.")
                self:ChangeButtonTarget(found)
                return true
            end
        end
    end
end

--[[
    SkillInfo
        Updates the list of skills the player has and their respective skill
        levels.
--]]
function Expeller:SkillInfo()
    local sR, sN
    wipe(self.craft_skill)
    local prof1, prof2 = GetProfessions()
    if prof1 ~= nil then
        local sN, _, sR = GetProfessionInfo(prof1)
        if self.lookup[sN] then 
            self.craft_skill[sN] = sR
        end
    end
    if prof2 ~= nil then
        local sN, _, sR = GetProfessionInfo(prof2)
        if self.lookup[sN] then 
            self.craft_skill[sN] = sR
        end
    end
    return sR, sN
end
